# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

import unittest

from slo_generator.backends.splunk import SplunkBackend


class TestSplunkBackend(unittest.TestCase):
    def test_fix_search_prefix(search):
        search: str = "index=* status=200"
        fixed_search = "search " + search

        assert SplunkBackend.fix_search_prefix(search) == fixed_search
        assert SplunkBackend.fix_search_prefix(fixed_search) == fixed_search
